/* mmc.c - mmap cache
**
** Copyright  1998,2001 by Jef Poskanzer <jef@acme.com>.
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>

#ifdef HAVE_MMAP
#include <sys/mman.h>
#endif /* HAVE_MMAP */

#include "mmc.h"


/* Macro */
#ifndef DESIRED_TO_SW_LIMIT
#define DESIRED_TO_SW_LIMIT(n)		( ( (n) * 5 ) / 4 )
#endif

/* Defines */
#ifndef DEFAULT_CLEANUP_TIME
#define DEFAULT_CLEANUP_TIME	10
#endif
#ifndef DEFAULT_EXPIRE_AGE
#define DEFAULT_EXPIRE_AGE	300
#endif
#ifndef DEFAULT_EXPIRE_AGE_L0
#define DEFAULT_EXPIRE_AGE_L0()	( cfgDefaultExpireAge * 16 )
#endif
#ifndef DESIRED_FREE_MAC_COUNT
#define DESIRED_FREE_MAC_COUNT	500
#endif
#ifndef DESIRED_FREE_MMC_COUNT
#define DESIRED_FREE_MMC_COUNT	100
#endif
#ifndef DESIRED_FREE_FDC_COUNT
#define DESIRED_FREE_FDC_COUNT	100
#endif
#ifndef DESIRED_MAX_MALLOC_FILES
#define DESIRED_MAX_MALLOC_FILES	10000
#endif
#ifndef DESIRED_MAX_MALLOC_BYTES
#define DESIRED_MAX_MALLOC_BYTES	1000000
#endif
#ifndef DESIRED_MAX_MAPPED_FILES
#define DESIRED_MAX_MAPPED_FILES	2000
#endif
#ifndef DESIRED_MAX_OPENED_FILES
#define DESIRED_MAX_OPENED_FILES	100
#endif
#ifndef MAX_FILE_SIZE_L0
#define MAX_FILE_SIZE_L0	(0)
#endif
#ifndef MAX_FILE_SIZE_L1
#define MAX_FILE_SIZE_L1	(1024*256)
#endif
#ifndef MAX_FILE_SIZE_L2
#define MAX_FILE_SIZE_L2	(1024*1024)
#endif
#ifndef MAX_FILE_SIZE_L3
#define MAX_FILE_SIZE_L3	(1024*1024*4)
#endif

#ifndef INITIAL_HASH_BITS
#define INITIAL_HASH_BITS	10
#endif
#ifndef INITIAL_HASH_SIZE
#define INITIAL_HASH_SIZE	(1 << INITIAL_HASH_BITS)
#endif
#ifndef AVG_HASH_CHAIN_LENGTH
#define AVG_HASH_CHAIN_LENGTH	4
#endif

#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)<(b)?(a):(b))

/* defines for mmc_unmap_[un]ref() */
#define NO_LOG	0
#define DO_LOG	1

/* defines for mmc_ma_cleanup() */
#define NO_AGE	0
#define DO_AGE	1
#define NO_FREE	0
#define DO_FREE	1

/* cache types */
#define MMC_CACHE_MAMAP		1
#define MMC_CACHE_MMMAP		2
#define MMC_CACHE_FDMAP		3

/* macro for mmc_is_fdmap() */
#define MMC_IS_FDMAP( fsize )	( ( fsize ) > cfgMaxFileSizeL3 && cfgUseFdMap )

/* The Map struct. */
typedef struct MapStruct {
    ino_t  ino;
    dev_t  dev;
    off_t  size;
    time_t ctime;
    int    refcount;
    time_t reftime;
    int    fd;
    void*  addr;
    unsigned
    int    hash;
    struct MapStruct* prev_lru;
    struct MapStruct* next_lru;
    struct MapStruct* prev_hash;
    struct MapStruct* next_hash;
    } Map;

typedef struct MapCtlStruct {
    int   cache_type;
    Map*  lru_ref;
    Map*  lru_unref;
    Map*  free_maps;
    Map** hash_table;
    int   hash_size;
    unsigned
    int   hash_mask;
    off_t alloc_bytes;
    int   alloc_count;
    int   map_count;
    int   use_count;
    int   free_count;
} MapCtl;

/* Globals. */

static MapCtl maMapCtl =
    {
	MMC_CACHE_MAMAP
    };
static MapCtl mmMapCtl =
    {
	MMC_CACHE_MMMAP
    };
static MapCtl fdMapCtl =
    {
	MMC_CACHE_FDMAP
    };

static time_t expire_age_L1 = DEFAULT_EXPIRE_AGE;
static time_t expire_age_L2 = DEFAULT_EXPIRE_AGE / 2;
static time_t expire_age_L3 = DEFAULT_EXPIRE_AGE / 8;

static int cfgDefaultExpireAge = DEFAULT_EXPIRE_AGE;
static int cfgCleanupTime      = DEFAULT_CLEANUP_TIME;
static off_t cfgDesiredMaxMallocBytes = DESIRED_MAX_MALLOC_BYTES;
static off_t cfgSwLimitMaxMallocBytes =
		 DESIRED_TO_SW_LIMIT( DESIRED_MAX_MALLOC_BYTES );
static int cfgDesiredMaxMallocFiles = DESIRED_MAX_MALLOC_FILES;
static int cfgSwLimitMaxMallocFiles =
		 DESIRED_TO_SW_LIMIT( DESIRED_MAX_MALLOC_FILES );
static int cfgDesiredMaxMappedFiles = DESIRED_MAX_MAPPED_FILES;
static int cfgDesiredMaxOpenedFiles = DESIRED_MAX_OPENED_FILES;
static int cfgUseFdMap = 0;
static int cfgCloseOnExec = 0;
static size_t cfgMaxFileSizeL0 = MAX_FILE_SIZE_L0;
static size_t cfgMaxFileSizeL1 = MAX_FILE_SIZE_L1;
static size_t cfgMaxFileSizeL2 = MAX_FILE_SIZE_L2;
static size_t cfgMaxFileSizeL3 = MAX_FILE_SIZE_L3;

/* Forwards. */
static inline void add_lru_ref( MapCtl* const pmc, Map* const m );
static inline void add_lru_unref( MapCtl* const pmc, Map* const m );
static inline void del_lru( Map* const m );
static inline unsigned int hash_fstat( unsigned int hash_mask,
			ino_t ino, dev_t dev, off_t size, time_t ctime );
static inline void add_hash( MapCtl* const pmc, Map* m );
static inline void del_hash( MapCtl* const pmc, Map* m );

static int  check_hash_lru_size( MapCtl* const pmc);
static Map* find_hash( MapCtl* const pmc,
			ino_t ino, dev_t dev, off_t size, time_t ctime );
static void mmc_really_unmap( MapCtl* const pmc, Map* mm );
static void mmc_unmap_unref( MapCtl* const pmc, const int do_log );
static void mmc_unmap_ref( MapCtl* const pmc, const int do_log );
static int  mmc_cache_type( off_t st_size );
static inline
	int mmc_pmc_cache_type( MapCtl** ppmc, off_t st_size );
static void mmc_pmc_free( MapCtl* const pmc, const int min_count );
static void mmc_ma_cleanup( struct timeval* nowP, int do_age, int do_free );
static void mmc_mm_cleanup( struct timeval* nowP );
static void mmc_fd_cleanup( void );
static void mmc_pmc_destroy( MapCtl* const pmc );


static inline void
add_lru_ref( MapCtl* const pmc, Map* const m )
    {
    m->prev_lru = pmc->lru_ref->prev_lru;
    m->next_lru = pmc->lru_ref;
    pmc->lru_ref->prev_lru->next_lru = m;
    pmc->lru_ref->prev_lru = m;
    }


static inline void
add_lru_unref( MapCtl* const pmc, Map* const m )
    {
    m->prev_lru = pmc->lru_unref->prev_lru;
    m->next_lru = pmc->lru_unref;
    pmc->lru_unref->prev_lru->next_lru = m;
    pmc->lru_unref->prev_lru = m;
    }


static inline void
del_lru( Map* const m )
    {
    m->prev_lru->next_lru = m->next_lru;
    m->next_lru->prev_lru = m->prev_lru;
    }


static inline unsigned int
hash_fstat( unsigned int hash_mask,
	ino_t ino, dev_t dev, off_t size, time_t ctime )
    {
    unsigned int h = 177573;

    h ^= ino;
    h += h << 5;
    h ^= dev;
    h += h << 5;
    h ^= size;
    h += h << 5;
    h ^= ctime;

    return h & hash_mask;
    }


static inline void
add_hash( MapCtl* const pmc, Map* m )
    {
    m->hash = hash_fstat( pmc->hash_mask, m->ino, m->dev, m->size, m->ctime );
    if ( pmc->hash_table[m->hash] == NULL )
	{
	pmc->hash_table[m->hash] = m;
	m->prev_hash = m;
	m->next_hash = m;
	}
    else
	{
	Map *p = pmc->hash_table[m->hash];
	m->prev_hash = p->prev_hash;
	m->next_hash = p;
	p->prev_hash->next_hash = m;
	p->prev_hash = m;
	}
    }


static inline void
del_hash( MapCtl* const pmc, Map* m )
    {
    if ( pmc->hash_table[m->hash] == m )
	{
	if ( m->next_hash == m )
	    {  /* there is only this node on current hash chain */
	    pmc->hash_table[m->hash] = NULL;
	    return;
	    }
	pmc->hash_table[m->hash] = m->next_hash;
	}
    m->prev_hash->next_hash = m->next_hash;
    m->next_hash->prev_hash = m->prev_hash;
    }


/* Make sure the hash table is big enough. */
static int
check_hash_lru_size( MapCtl* const pmc )
    {
    Map* m;

    /* Are we just starting out? */
    if ( pmc->hash_table == (Map**) 0 )
	{
	pmc->hash_size = INITIAL_HASH_SIZE;
	pmc->hash_mask = pmc->hash_size - 1;

	if ( pmc->lru_ref == NULL )
	    {
	    /* allocate sentinel node for referenced map list */
	    pmc->lru_ref = (Map*) malloc( sizeof(Map) );
	    if ( pmc->lru_ref == (Map*) 0 )
		return MMC_ERROR;
	    memset( pmc->lru_ref, 0, sizeof(Map) );
	    pmc->lru_ref->prev_lru = pmc->lru_ref;
	    pmc->lru_ref->next_lru = pmc->lru_ref;
	    }

	if ( pmc->lru_unref == NULL )
	    {
	    /* allocate sentinel node for unreferenced map list */
	    pmc->lru_unref = (Map*) malloc( sizeof(Map) );
	    if ( pmc->lru_unref == (Map*) 0 )
		{ /* free previous sentinel node */
		free( pmc->lru_ref );
		pmc->lru_ref = NULL;
		return MMC_ERROR;
		}
	    memset( pmc->lru_unref, 0, sizeof(Map) );
	    pmc->lru_unref->prev_lru = pmc->lru_unref;
	    pmc->lru_unref->next_lru = pmc->lru_unref;
	    }
	}
    /* Is it at least three times bigger than the number of entries? */
    else if ( pmc->hash_size * AVG_HASH_CHAIN_LENGTH >= pmc->map_count )
	return MMC_NORMAL;
    else
	{
	/* No, got to expand. */
	free( (void*) pmc->hash_table );
	/* Double the hash size until it's big enough. */
	do
	    {
	    pmc->hash_size = pmc->hash_size << 1;
	    }
	while ( pmc->hash_size * AVG_HASH_CHAIN_LENGTH < pmc->map_count );
	pmc->hash_mask = pmc->hash_size - 1;
	}

    /* Make the new table. */
    pmc->hash_table = (Map**) calloc( pmc->hash_size, sizeof(Map*) );
    if ( pmc->hash_table == (Map**) 0 )
	return MMC_ERROR;

    /* (already cleared by calloc) */

    /* And rehash all entries. */
    for ( m = pmc->lru_ref->next_lru; m != pmc->lru_ref; m = m->next_lru )
	add_hash( pmc, m );

    for ( m = pmc->lru_unref->next_lru; m != pmc->lru_unref; m = m->next_lru )
	add_hash( pmc, m );

    return MMC_NORMAL;
    }


static Map*
find_hash( MapCtl* const pmc, ino_t ino, dev_t dev, off_t size, time_t ctime )
    {
    unsigned int h;
    Map* m0;
    Map* m;

    h = hash_fstat( pmc->hash_mask, ino, dev, size, ctime );
    m = m0 = pmc->hash_table[h];
    if ( m0 == NULL )
	return (Map*) 0;
    do
	{
	if ( m->ino == ino && m->dev == dev &&
	     m->size == size && m->ctime == ctime )
	    return m;
	m = m->next_hash;
	}
    while ( m != m0 );

    return (Map*) 0;
    }


static void
mmc_really_unmap( MapCtl* const pmc, Map* m )
    {

    if ( m->fd != EOF )
	{
	(void) close( m->fd );
	}
    else
    if ( m->size != 0 )
	{
#ifdef HAVE_MMAP
	if ( pmc->cache_type == MMC_CACHE_MMMAP )
	    {
	    if ( munmap( m->addr, m->size ) < 0 )
		syslog( LOG_ERR, "munmap - %m" );
	    }
	else
	    {   /* MMC_CACHE_MAMAP */
#endif /* HAVE_MMAP */
	    pmc->alloc_bytes -= m->size;
	    free( (void*) m->addr );
	    m->addr = NULL;
#ifdef HAVE_MMAP
	    }
#endif /* HAVE_MMAP */
	}

    /* Detach from unreferenced list */
    del_lru( m );

    /* Detach from hash list */
    del_hash( pmc, m );

    /* and move the Map to the free list. */
    --pmc->map_count;
    if ( m->refcount > 0 )
	--pmc->use_count;
    m->next_lru = pmc->free_maps;
    pmc->free_maps = m;
    ++pmc->free_count;
    }


static void
mmc_unmap_unref( MapCtl* const pmc, const int do_log )
    {
    Map* m0;
    Map* m;

    if ( do_log )
	{
	char	*pszErrno;
	if ( pmc->cache_type == MMC_CACHE_FDMAP )
	    pszErrno = "EMFILE";
	else
	    pszErrno = "ENOMEM";
	syslog( LOG_WARNING, "mmc: %s - freeing all unreferenced maps (%d)",
	    pszErrno, ( pmc->map_count - pmc->use_count ) );
	}

    /* check if something has been allocated */
    if ( pmc->lru_unref == (Map*) 0 )
	return;

    /* Really unmap all unreferenced entries. */
    for ( m = pmc->lru_unref->next_lru; m != pmc->lru_unref; )
	{
	m0 = m;
	m = m->next_lru;
	mmc_really_unmap( pmc, m0 );
	}
    pmc->lru_unref->prev_lru = pmc->lru_unref;
    pmc->lru_unref->next_lru = pmc->lru_unref;

    /* no free */
    }


static void
mmc_unmap_ref( MapCtl* const pmc, const int do_log )
    {
    Map* m0;
    Map* m;

    if ( do_log )
	{
	char	*pszErrno;
	if ( pmc->cache_type == MMC_CACHE_FDMAP )
	    pszErrno = "EMFILE";
	else
	    pszErrno = "ENOMEM";
	syslog( LOG_WARNING, "mmc: %s - freeing all referenced maps (%d)",
	    pszErrno, pmc->use_count );
	}

    /* check if something has been allocated */
    if ( pmc->lru_ref == (Map*) 0 )
	return;

    /* Really unmap all referenced entries. */
    for ( m = pmc->lru_ref->next_lru; m != pmc->lru_ref; )
	{
	m0 = m;
	m = m->next_lru;
	mmc_really_unmap( pmc, m0 );
	}
    pmc->lru_ref->prev_lru = pmc->lru_ref;
    pmc->lru_ref->next_lru = pmc->lru_ref;

    /* no free */
    }


int
mmc_get_value( int value_type )
    {
	switch( value_type )
	{
	case MMC_V_MA_MAP_COUNT:
		return maMapCtl.map_count;
	case MMC_V_MA_USE_COUNT:
		return maMapCtl.use_count;
	case MMC_V_MM_MAP_COUNT:
		return mmMapCtl.map_count;
	case MMC_V_MM_USE_COUNT:
		return mmMapCtl.use_count;
	case MMC_V_FD_MAP_COUNT:
		return fdMapCtl.map_count;
	case MMC_V_FD_USE_COUNT:
		return fdMapCtl.use_count;
	default:
		return -1;
	}
    }


int
mmc_cfg_get_param( int param_type )
    {
	switch( param_type )
	{
	case MMC_P_CLEANUP_TIME:
		return cfgCleanupTime;
	case MMC_P_DEFAULT_EXPIRE_AGE:
		return cfgDefaultExpireAge;
	case MMC_P_DESIRED_MAX_MALLOC_BYTES:
		return (int) cfgDesiredMaxMallocBytes;
	case MMC_P_DESIRED_MAX_MALLOC_FILES:
		return cfgDesiredMaxMallocFiles;
	case MMC_P_DESIRED_MAX_MAPPED_FILES:
		return cfgDesiredMaxMappedFiles;
	case MMC_P_DESIRED_MAX_OPENED_FILES:
		return cfgDesiredMaxOpenedFiles;
	case MMC_P_USE_SENDFILE:
		return cfgUseFdMap;
	case MMC_P_CLOSE_ON_EXEC:
		return cfgCloseOnExec;
	case MMC_P_MAX_FILE_SIZE_L0:
		return (int) cfgMaxFileSizeL0;
	case MMC_P_MAX_FILE_SIZE_L1:
		return (int) cfgMaxFileSizeL1;
	case MMC_P_MAX_FILE_SIZE_L2:
		return (int) cfgMaxFileSizeL2;
	case MMC_P_MAX_FILE_SIZE_L3:
		return (int) cfgMaxFileSizeL3;
	default:
		return -1;
	}
    }


int
mmc_cfg_set_param( int param_type, int param_value )
    {
	switch( param_type )
	{
	case MMC_P_CLEANUP_TIME:
		cfgCleanupTime = param_value;
		if ( cfgCleanupTime < 2 )
		     cfgCleanupTime = 2;
		else
		if ( cfgCleanupTime > 200 )
		     cfgCleanupTime = 200;
		param_value = cfgCleanupTime;
		break;

	case MMC_P_DEFAULT_EXPIRE_AGE:
		if ( param_value < 10 )
		     param_value = 10;
		else
		if ( param_value > 10000 )
		     param_value = 10000;
		expire_age_L1 = param_value;
		cfgDefaultExpireAge = param_value;
		break;

	case MMC_P_DESIRED_MAX_MALLOC_BYTES:
		if ( param_value < 1000 )
		     param_value = 1000;
		else
		if ( param_value > 100000000 )
		     param_value = 100000000;
		cfgDesiredMaxMallocBytes = (off_t) param_value;
		cfgSwLimitMaxMallocBytes = (off_t)
				DESIRED_TO_SW_LIMIT( param_value );
		break;

	case MMC_P_DESIRED_MAX_MALLOC_FILES:
		if ( param_value < 2 )
		     param_value = 2;
		else
		if ( param_value > 500000 )
		     param_value = 500000;
		cfgDesiredMaxMallocFiles = param_value;
		cfgSwLimitMaxMallocFiles = DESIRED_TO_SW_LIMIT( param_value );
		break;

	case MMC_P_DESIRED_MAX_MAPPED_FILES:
		if ( param_value < 2 )
		     param_value = 2;
		else
		if ( param_value > 40000 )
		     param_value = 40000;
		cfgDesiredMaxMappedFiles = param_value;
		break;

	case MMC_P_DESIRED_MAX_OPENED_FILES:
		if ( param_value < 10 )
		     param_value = 10;
		else
		if ( param_value > 4000 )
		     param_value = 4000;
		cfgDesiredMaxOpenedFiles = param_value;
		break;

	case MMC_P_USE_SENDFILE:
		cfgUseFdMap = !!param_value;
		break;

	case MMC_P_CLOSE_ON_EXEC:
		cfgCloseOnExec = !!param_value;
		break;

	case MMC_P_MAX_FILE_SIZE_L0:
		if ( param_value < 0 )
		     param_value = 0;
		else
		if ( param_value > 4096 )
		     param_value = 4096;
		cfgMaxFileSizeL0 = (size_t) param_value;
		if ( cfgMaxFileSizeL0 > cfgMaxFileSizeL1)
		     cfgMaxFileSizeL1 = cfgMaxFileSizeL0 + 1024;
		if ( cfgMaxFileSizeL1 > cfgMaxFileSizeL2)
		     cfgMaxFileSizeL2 = cfgMaxFileSizeL1 + 1024;
		if ( cfgMaxFileSizeL2 > cfgMaxFileSizeL3)
		     cfgMaxFileSizeL3 = cfgMaxFileSizeL2 + 1024;
		param_value = (int) cfgMaxFileSizeL0;
		break;

	case MMC_P_MAX_FILE_SIZE_L1:
		if ( param_value < 32 )
		     param_value = 32;
		else
		if ( param_value > (1024*1024*1024) )
		     param_value = (1024*1024*1024);
		cfgMaxFileSizeL1 = (size_t) param_value;
		if ( cfgMaxFileSizeL0 > cfgMaxFileSizeL1)
		     cfgMaxFileSizeL1 = cfgMaxFileSizeL0 + 1024;
		if ( cfgMaxFileSizeL1 > cfgMaxFileSizeL2)
		     cfgMaxFileSizeL2 = cfgMaxFileSizeL1 + 1024;
		if ( cfgMaxFileSizeL2 > cfgMaxFileSizeL3)
		     cfgMaxFileSizeL3 = cfgMaxFileSizeL2 + 1024;
		param_value = (int) cfgMaxFileSizeL1;
		break;

	case MMC_P_MAX_FILE_SIZE_L2:
		cfgMaxFileSizeL2 = (size_t) param_value;
		if ( cfgMaxFileSizeL2 <= cfgMaxFileSizeL1 )
		     cfgMaxFileSizeL2  = cfgMaxFileSizeL1 + 1024;
		if ( cfgMaxFileSizeL3 <= cfgMaxFileSizeL2 )
		     cfgMaxFileSizeL3  = cfgMaxFileSizeL2 + 1024;
		param_value = (int) cfgMaxFileSizeL2;
		break;

	case MMC_P_MAX_FILE_SIZE_L3:
		cfgMaxFileSizeL3 = (size_t) param_value;
		if ( cfgMaxFileSizeL3 <= cfgMaxFileSizeL1 )
		     cfgMaxFileSizeL3  = cfgMaxFileSizeL2 + 1024;
		if ( cfgMaxFileSizeL3 <= cfgMaxFileSizeL2 )
		     cfgMaxFileSizeL3  = cfgMaxFileSizeL2 + 1024;
		param_value = (int) cfgMaxFileSizeL3;
		break;

	default:
		return -1;
	}
	return param_value;
    }


int
mmc_is_fdmap( off_t st_size )
    {
    return ( MMC_IS_FDMAP( st_size ) );
    }


static int
mmc_cache_type( off_t st_size )
    {
    if ( st_size <  0 )
	return( MMC_ERROR );
    if ( st_size <= cfgMaxFileSizeL0 )
	return ( MMC_CACHE_MAMAP );
    if ( MMC_IS_FDMAP( st_size ) )
	return ( MMC_CACHE_FDMAP );
#ifdef HAVE_MMAP
    return ( MMC_CACHE_MMMAP );
#else
    return ( MMC_CACHE_MAMAP );
#endif
    }


static inline int
mmc_pmc_cache_type( MapCtl** ppmc, off_t st_size )
    {
    int cache_type = mmc_cache_type( st_size );
    switch( cache_type )
	{
	case MMC_CACHE_MAMAP:
	    *ppmc = &maMapCtl;
	    break;
	case MMC_CACHE_MMMAP:
	    *ppmc = &mmMapCtl;
	    break;
	case MMC_CACHE_FDMAP:
	    *ppmc = &fdMapCtl;
	    break;
	default:
	    return( MMC_ERROR );
	}
    return( cache_type );
    }


int
mmc_map( int* pfd, void** paddr,
	char* filename, struct stat* sbP, struct timeval* nowP )
    {
    time_t now;
    struct stat sb;
    Map* m;
    MapCtl *pmc;
    int fd;

    /* Stat the file, if necessary. */
    if ( sbP != (struct stat*) 0 )
	sb = *sbP;
    else
	{
	if ( stat( filename, &sb ) != 0 )
	    {
	    syslog( LOG_ERR, "mmc: stat - %m" );
	    return MMC_ERROR;
	    }
	}

    /* Get the current time, if necessary. */
    if ( nowP != (struct timeval*) 0 )
	now = nowP->tv_sec;
    else
	now = time( (time_t*) 0 );

    if ( mmc_pmc_cache_type( &pmc, sb.st_size ) == MMC_ERROR )
	{
	syslog( LOG_ERR, "mmc: map pmc_cache, st_size %ld", sb.st_size );
	return MMC_ERROR;
	}

    /* See if we have it mapped already, via the hash table. */
    if ( check_hash_lru_size( pmc ) != MMC_NORMAL )
	{
	syslog( LOG_ERR, "mmc: check_hash_lru_size() failure" );
	return MMC_ERROR;
	}

    m = find_hash( pmc, sb.st_ino, sb.st_dev, sb.st_size, sb.st_ctime );

    if ( m != (Map*) 0 )
	{
	/* Yep.  Just return the existing map */
	if ( m->refcount == 0 )
	    {
	    del_lru( m );
	    add_lru_ref( pmc, m );
	    ++pmc->use_count;
	    }
	++m->refcount;
	m->reftime = now;
	*pfd = m->fd;
	*paddr = m->addr;
	return MMC_NORMAL;
	}

    /* Open the file (we could avoid this for size == 0) */
    fd = open( filename, O_RDONLY );
    if ( fd < 0 )
	{
	if ( pmc->cache_type == MMC_CACHE_FDMAP &&
	    ( errno == EMFILE || errno == ENFILE ) &&
	    pmc->use_count < pmc->map_count )
	    {
	    int errno0 = errno;
	    /* free unused maps (for next time) */
	    mmc_unmap_unref( pmc, DO_LOG );
	    errno = errno0;
	    }
	syslog( LOG_ERR, "mmc: open - %m" );
	return MMC_ERROR;
	}

    /* Find a free Map entry or make a new one. */
    if ( pmc->free_maps != (Map*) 0 )
	{
	m = pmc->free_maps;
	pmc->free_maps = m->next_lru;
	--pmc->free_count;
	}
    else
	{
	m = (Map*) malloc( sizeof(Map) );
	if ( m == (Map*) 0 )
	    {
	    (void) close( fd );
	    syslog( LOG_ERR, "mmc: out of memory allocating a Map" );
	    return MMC_ERROR;
	    }
	++pmc->alloc_count;
	}

    /* Fill in the Map entry. */
    m->ino = sb.st_ino;
    m->dev = sb.st_dev;
    m->size = sb.st_size;
    m->ctime = sb.st_ctime;
    m->refcount = 1;
    m->reftime = now;

    if ( pmc->cache_type == MMC_CACHE_FDMAP )
	{
	m->fd = fd;
	m->addr = (void*) 0;
	if ( cfgCloseOnExec )
	    {
	    /* set close-on-exec flag (useful for spawned subprocesses) */
	    if ( fcntl( fd, F_SETFD, 1 ) != 0 )
		{
		syslog( LOG_ERR, "mmc: fcntl - %m" );
		(void) close( fd );
		free( (void*) m );
		--pmc->alloc_count;
		return MMC_ERROR;
		}
	    }
	}
    else
	{
	m->fd = EOF;

	/* Avoid doing anything for zero-length files;
	** some systems don't like to mmap them,
	** other systems dislike mallocing zero bytes.
	*/
	if ( m->size == 0 )
	    m->addr = (void*) 1;	/* arbitrary non-NULL address */
	else
	    {
#ifdef HAVE_MMAP
	    if ( pmc->cache_type == MMC_CACHE_MMMAP )
		{
		/* Map the file into memory. */
		m->addr = mmap( 0, m->size, PROT_READ, MAP_SHARED, fd, 0 );
		if ( m->addr == (void*) -1 && errno == ENOMEM &&
		    pmc->use_count < pmc->map_count )
		    {
		    /* Out of address space, we have some unreferenced maps,
		    ** thus free all of them and try again.
		    */
		    mmc_unmap_unref( pmc, DO_LOG );
		    m->addr = mmap( 0, m->size, PROT_READ, MAP_SHARED, fd, 0 );
		    }
		if ( m->addr == (void*) -1 )
		    {
		    syslog( LOG_ERR, "mmc: mmap - %m (file size %ld)",
			m->size );
		    (void) close( fd );
		    free( (void*) m );
		    --pmc->alloc_count;
		    return MMC_ERROR;
		    }
		}
	    else
		{   /* pmc->cache_type == MMC_CACHE_MAMAP */
#endif /* HAVE_MMAP */
		/* Read the file into memory. */
		m->addr = (void*) malloc( m->size );
		if ( m->addr == (void*) 0 && errno == ENOMEM &&
		    pmc->use_count < pmc->map_count )
		    {
		    /* Out of address space, we have some unreferenced maps,
		    ** thus free all of them and try again.
		    */
		    mmc_unmap_unref( pmc, DO_LOG );
		    m->addr = (void*) malloc( m->size );
		}
		if ( m->addr == (void*) 0 )
		    {
		    syslog( LOG_ERR, "mmc: out of memory (file size %ld)",
			m->size );
		    (void) close( fd );
		    free( (void*) m );
		    --pmc->alloc_count;
		    return MMC_ERROR;
		    }
		if ( read( fd, m->addr, m->size ) != m->size )
		    {
		    syslog( LOG_ERR, "mmc: read - %m" );
		    (void) close( fd );
		    free( (void*) m );
		    --pmc->alloc_count;
		    return MMC_ERROR;
		    }
		pmc->alloc_bytes += m->size;
#ifdef HAVE_MMAP
		}
#endif /* HAVE_MMAP */
	    }
	(void) close( fd );
	}

    /* Put the Map into the hash table. */
    add_hash( pmc, m );

    /* Put the Map on the lru referenced list. */
    add_lru_ref( pmc, m );

    ++pmc->map_count;
    ++pmc->use_count;

    /* Assign values to formal parameters */
    *pfd = m->fd;
    *paddr = m->addr;

    /* And return OK */
    return MMC_NORMAL;
    }


int
mmc_unmap( int fd, void* addr, struct stat* sbP, struct timeval* nowP )
    {
    Map* m = (Map*) 0;
    MapCtl *pmc;

    if ( mmc_pmc_cache_type( &pmc, sbP->st_size ) == MMC_ERROR )
	{
	syslog( LOG_ERR, "mmc: map pmc_cache, st_size %ld", sbP->st_size );
	return MMC_ERROR;
	}

    /* Find the Map entry for this address.  First try a hash. */
    m = find_hash( pmc, sbP->st_ino, sbP->st_dev, sbP->st_size,
			sbP->st_ctime );
    if ( pmc->cache_type == MMC_CACHE_FDMAP )
	{
	if ( m != (Map*) 0 && m->fd != fd )
	    {  /* fatal error */
	    syslog( LOG_ERR, "mmc_unmap: found bad entry fd %d != %d !",
		m->fd, fd );
	    return MMC_ERROR;
	    }
	if ( m == (Map*) 0 )
	    {  /* fatal error */
	    syslog( LOG_ERR, "mmc_unmap: entry fd %d not found !",
		fd );
	    return MMC_ERROR;
	    }
	}
    else
	{   /* MMC_CACHE_MAMAP || MMC_CACHE_MMMAP */
	if ( m != (Map*) 0 && m->addr != addr )
	    {  /* fatal error */
	    syslog( LOG_ERR, "mmc_unmap: found bad entry addr %p != %p !",
		m->addr, addr );
	    return MMC_ERROR;
	    }
	if ( m == (Map*) 0 )
	    {  /* fatal error */
	    syslog( LOG_ERR, "mmc_unmap: entry addr %p not found !",
		addr );
	    return MMC_ERROR;
	    }
	}

    if ( m->refcount <= 0 )
	{
	syslog( LOG_ERR, "mmc_unmap found zero or negative refcount!" );
	return MMC_ERROR;
	}

    if ( nowP != (struct timeval*) 0 )
	m->reftime = nowP->tv_sec;
    else
	m->reftime = time( (time_t*) 0 );

    if ( --m->refcount == 0 )
	{
	/* remove this entry from referenced and add it to unreferenced list */
	del_lru( m );
	add_lru_unref( pmc, m );
	--pmc->use_count;

	/* eventually shrink cache */
	switch( pmc->cache_type )
	    {
	    case MMC_CACHE_MAMAP:
		if ( pmc->map_count   > cfgSwLimitMaxMallocFiles ||
		     pmc->alloc_bytes > cfgSwLimitMaxMallocBytes )
		     mmc_ma_cleanup( nowP, NO_AGE, NO_FREE );
		break;
	    case MMC_CACHE_MMMAP:
		/* nothing for now */
		break;
	    case MMC_CACHE_FDMAP:
		if ( pmc->map_count >
		     pmc->use_count + cfgDesiredMaxOpenedFiles )
		     mmc_fd_cleanup( );
		break;
	    default:
		break;
	    }
	}

    /* OK, return */
    return MMC_NORMAL;
    }


static void
mmc_pmc_free( MapCtl* const pmc, const int min_count )
    {
    Map* m;
    /* Really free excess blocks on the free list. */
    while ( pmc->free_maps != (Map*) 0 &&
	pmc->free_count > min_count )
	{
	m = pmc->free_maps;
	pmc->free_maps = m->next_lru;
	--pmc->free_count;
	free( (void*) m );
	--pmc->alloc_count;
	}
    }


static void
mmc_ma_cleanup( struct timeval* nowP, int do_age, int do_free )
    {
    time_t now;
    Map* m0;
    Map* m;
    MapCtl* const pmc = &maMapCtl;

    /* Get the current time, if necessary. */
    if ( nowP != (struct timeval*) 0 )
	now = nowP->tv_sec;
    else
	now = time( (time_t*) 0 );

    /* check if something has been allocated */
    if ( pmc->lru_unref == (Map*) 0 )
	return;

    /* Really unmap any unreferenced entries until we go under file limit. */
    if ( pmc->map_count > cfgDesiredMaxMallocFiles )
	{
	int	unref_maps   = pmc->map_count - pmc->use_count;
	int	maps_to_free = pmc->map_count - cfgDesiredMaxMallocFiles;

	if ( unref_maps < 1 )
	    return;

	if ( maps_to_free > unref_maps )
	     maps_to_free = unref_maps;
	for ( m = pmc->lru_unref->next_lru;
	      m != pmc->lru_unref && maps_to_free > 0;
	      --maps_to_free )
	    {
	    m0 = m;
	    m = m->next_lru;
	    mmc_really_unmap( pmc, m0 );
	    }
	}

    /* Really unmap any unreferenced entries until we go under byte limit. */
    if ( pmc->alloc_bytes > cfgDesiredMaxMallocBytes )
	{
	for ( m = pmc->lru_unref->next_lru;
	      m != pmc->lru_unref &&
	      pmc->alloc_bytes > cfgDesiredMaxMallocBytes;
	    )
	    {
	    m0 = m;
	    m = m->next_lru;
	    mmc_really_unmap( pmc, m0 );
	    }
	}

    /* Really unmap any unreferenced entries older than the age limit. */
    if ( do_age )
	{
	if ( now <= DEFAULT_EXPIRE_AGE_L0() )
	    now = 0;
	else
	    now -= DEFAULT_EXPIRE_AGE_L0();

	for ( m = pmc->lru_unref->next_lru;
	      m != pmc->lru_unref &&
	      m->reftime < now;
	    )
	    {
	    m0 = m;
	    m = m->next_lru;
	    mmc_really_unmap( pmc, m0 );
	    }
	}

    /* Really free excess blocks on the free list. */
    if ( do_free )
	mmc_pmc_free( pmc, DESIRED_FREE_MAC_COUNT );
    }


static void
mmc_mm_cleanup( struct timeval* nowP )
    {
    time_t now;
    Map* m0;
    Map* m;
    MapCtl* const pmc = &mmMapCtl;
    time_t	cur_expire_age = 0;
    int		map_count0 = mmMapCtl.map_count;   /* count before cleanup */

    /* Get the current time, if necessary. */
    if ( nowP != (struct timeval*) 0 )
	now = nowP->tv_sec;
    else
	now = time( (time_t*) 0 );

    /* check if something has been allocated */
    if ( pmc->lru_unref == (Map*) 0 )
	return;

    /* Really unmap any unreferenced entries older than the age limit. */
    for ( m = pmc->lru_unref->next_lru; m != pmc->lru_unref; )
	{
	m0 = m;
	m = m->next_lru;

	if ( m0->size <= cfgMaxFileSizeL1 )
	    cur_expire_age = expire_age_L1;
	else
	if ( m0->size <= cfgMaxFileSizeL2 )
	    cur_expire_age = expire_age_L2;
	else
	if ( m0->size <= cfgMaxFileSizeL3 )
	    cur_expire_age = expire_age_L3;
	else
	    cur_expire_age = 0;

	if (  now < m0->reftime ||
	    ( now - m0->reftime ) >= cur_expire_age )
	    mmc_really_unmap( pmc, m0 );
	}

    /* If there are still too many (or too few) maps, adjust the age limit. */
    if ( pmc->map_count > cfgDesiredMaxMappedFiles )
	{	/* > 100% */
	if ( pmc->map_count < ( cfgDesiredMaxMappedFiles * 5 ) / 4 )
	    {	/* < 125% */
	    if ( expire_age_L1 > ( cfgDefaultExpireAge * 7 ) / 3 )
		expire_age_L1 = ( expire_age_L1 * 7 ) / 8;
	    else
	    if ( expire_age_L1 > cfgDefaultExpireAge )
		expire_age_L1 = ( expire_age_L1 * 5 ) / 6;
	    else
		expire_age_L1 =
		max( ( expire_age_L1 * 3 ) / 4, cfgDefaultExpireAge / 10 );
	    }
	else	/* >= 125% */
	    expire_age_L1 =
	    max( ( expire_age_L1 * 2 ) / 3, cfgDefaultExpireAge / 10 );
	}
    else if ( pmc->map_count < ( cfgDesiredMaxMappedFiles * 3 ) / 4 )
	{	/* < 75% */
	time_t	new_expire_age_L1;

	if ( map_count0 < 1 )
	    new_expire_age_L1 =
	    min( cfgDefaultExpireAge, cfgCleanupTime * 8 );
	else
	if ( expire_age_L1 > cfgDefaultExpireAge * 2 )
	    new_expire_age_L1 =
	    min( ( expire_age_L1 * 7 ) / 6, cfgDefaultExpireAge * 3 );
	else
	    new_expire_age_L1 =
	    min( ( expire_age_L1 * 5 ) / 4, cfgDefaultExpireAge * 3 );

	/* limit increase time to cleanup interval */
	/* NOTE: + 2 seconds so that we have some margin */
	/*	 for expire_age_L2 and expire_age_L3 */

	if ( expire_age_L1 + cfgCleanupTime < new_expire_age_L1 )
	     expire_age_L1 += cfgCleanupTime + 2;
	else
	     expire_age_L1 = new_expire_age_L1;
	}
    else if ( pmc->map_count > ( cfgDesiredMaxMappedFiles * 7 ) / 8 )
	{	/* > 87% */
	if ( expire_age_L1 >= cfgDefaultExpireAge * 3 )
	    expire_age_L1 = ( expire_age_L1 * 15 ) / 16;
	}
    /* else nothing, don't change current value of expire_age_L1 */

    expire_age_L2 = expire_age_L1 / 2;

    /* if thttpd is under heavy load (low expire_age_L1) */
    /* then don't decrease too much expire_age_L3 */

    if ( expire_age_L1 < cfgDefaultExpireAge / 8 )
	expire_age_L3 = expire_age_L1 / 4;
    else
    if ( expire_age_L1 < cfgDefaultExpireAge / 4 )
	expire_age_L3 = expire_age_L1 / 5;
    else
    if ( expire_age_L1 < cfgDefaultExpireAge / 2 )
	expire_age_L3 = expire_age_L1 / 6;
    else
    if ( expire_age_L1 < cfgDefaultExpireAge )
	expire_age_L3 = expire_age_L1 / 7;
    else
	expire_age_L3 = expire_age_L1 / 8;

    /* Really free excess blocks on the free list. */
    mmc_pmc_free( pmc, DESIRED_FREE_MMC_COUNT );
    }


static void
mmc_fd_cleanup( void )
    {
    Map* m0;
    Map* m;
    MapCtl* const pmc = &fdMapCtl;

    /* check if something has been allocated */
    if ( pmc->lru_unref == (Map*) 0 )
	return;

    /* Really unmap all unreferenced entries. */
    for ( m = pmc->lru_unref->next_lru; m != pmc->lru_unref; )
	{
	m0 = m;
	m = m->next_lru;
	mmc_really_unmap( pmc, m0 );
	}

    /* Really free excess blocks on the free list. */
    mmc_pmc_free( pmc, DESIRED_FREE_FDC_COUNT );
    }


void
mmc_cleanup( struct timeval* nowP )
    {
    mmc_ma_cleanup( nowP, DO_AGE, DO_FREE );
    mmc_mm_cleanup( nowP );
    mmc_fd_cleanup( );
    }


static void
mmc_pmc_destroy( MapCtl* const pmc )
    {
    mmc_unmap_unref( pmc, NO_LOG );
    mmc_unmap_ref( pmc, NO_LOG );
    mmc_pmc_free( pmc, 0 );
    }


void
mmc_destroy( void )
    {
    mmc_pmc_destroy( &maMapCtl );
    mmc_pmc_destroy( &mmMapCtl );
    mmc_pmc_destroy( &fdMapCtl );
    }


/* Generate debugging statistics syslog message. */
void
mmc_logstats( long secs )
    {
    MapCtl *pmc;

    pmc = &maMapCtl;

    syslog(
	LOG_NOTICE,
"  mal cache - %d allocated, %d active, %d used, %d free; hash size: %d; allocated bytes: %ld",
	pmc->alloc_count,
	pmc->map_count,
	pmc->use_count,
	pmc->free_count,
	pmc->hash_size,
	pmc->alloc_bytes );
    if ( pmc->map_count + pmc->free_count != pmc->alloc_count ||
	 pmc->use_count < 0 || pmc->use_count > pmc->map_count )
	syslog( LOG_ERR, "mmc: mal counts don't add up!" );

    pmc = &mmMapCtl;

    syslog(
	LOG_NOTICE,
"  map cache - %d allocated, %d active, %d used, %d free; hash size: %d; expire age L1: %ld",
	pmc->alloc_count,
	pmc->map_count,
	pmc->use_count,
	pmc->free_count,
	pmc->hash_size,
	expire_age_L1 );
    if ( pmc->map_count + pmc->free_count != pmc->alloc_count ||
	 pmc->use_count < 0 || pmc->use_count > pmc->map_count )
	syslog( LOG_ERR, "mmc: map counts don't add up!" );

    pmc = &fdMapCtl;

    syslog(
	LOG_NOTICE,
"  fd  cache - %d allocated, %d active, %d used, %d free; hash size: %d",
	pmc->alloc_count,
	pmc->map_count,
	pmc->use_count,
	pmc->free_count,
	pmc->hash_size );
    if ( pmc->map_count + pmc->free_count != pmc->alloc_count ||
	 pmc->use_count < 0 || pmc->use_count > pmc->map_count )
	syslog( LOG_ERR, "mmc: fd counts don't add up!" );
    }

